'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
require('source-map-support').install();
const program = require("commander");
const const_1 = require("../const");
const ProcessContext_1 = require("./ProcessContext");
const ProcfileReconciler_1 = require("./ProcfileReconciler");
const assert = require("assert");
const pandora_env_1 = require("pandora-env");
const LoggerBroker_1 = require("../universal/LoggerBroker");
const MonitorManager_1 = require("../monitor/MonitorManager");
const Facade_1 = require("../Facade");
const pandora_dollar_1 = require("pandora-dollar");
const ScalableMaster_1 = require("./ScalableMaster");
const SpawnWrapperUtils_1 = require("./SpawnWrapperUtils");
/**
 * class ProcessBootstrap
 * Bootstrap a worker process, handing all phases of an application stating
 */
class ProcessBootstrap {
    constructor(processRepresentation) {
        this.processRepresentation = processRepresentation;
        this.procfileReconciler = new ProcfileReconciler_1.ProcfileReconciler(processRepresentation);
        if (this.processRepresentation.scale > 1) {
            this.master = new ScalableMaster_1.ScalableMaster(processRepresentation);
            return;
        }
        this.context = new ProcessContext_1.ProcessContext(processRepresentation);
        Facade_1.Facade.set('processContext', this.context.processContextAccessor);
    }
    async start() {
        if (this.master) {
            await this.startAsMaster();
            return;
        }
        await this.startAsWorker();
    }
    async stop() {
        if (this.master) {
            await this.master.stop();
            return;
        }
        // Stop as worker
        await this.context.stop();
        SpawnWrapperUtils_1.SpawnWrapperUtils.unwrap();
    }
    async startAsMaster() {
        this.injectMonitor();
        await this.master.start();
    }
    async startAsWorker() {
        SpawnWrapperUtils_1.SpawnWrapperUtils.wrap();
        process.env[const_1.PANDORA_PROCESS] = JSON.stringify(this.processRepresentation);
        this.procfileReconciler.discover();
        this.injectMonitor();
        // Handing the services injecting
        const servicesByCurrentCategory = this.procfileReconciler.getServicesByCategory(this.processRepresentation.processName);
        this.context.bindService(servicesByCurrentCategory);
        // To start process by ProcessContext
        await this.context.start();
        // Require the entryFile if there given it, pandora.fork() dep on it
        if (this.processRepresentation.entryFile) {
            // A trick to make module === require.main be true, just remind here
            // debug('runMain pre', process.argv)
            // process.argv.splice(1, nargs)
            // process.argv[1] = path.resolve(process.argv[1])
            // delete require.cache[process.argv[1]]
            // debug('runMain post', process.argv)
            // Module.runMain()
            // debug('runMain after')
            const entryFileBaseDir = this.processRepresentation.entryFileBaseDir;
            const ownRequire = entryFileBaseDir ? pandora_dollar_1.makeRequire(entryFileBaseDir) : require;
            ownRequire(this.processRepresentation.entryFile);
        }
    }
    injectMonitor() {
        if (!pandora_env_1.EnvironmentUtil.getInstance().isReady()) {
            // Handing the environment object injecting
            const Environment = this.procfileReconciler.getEnvironment();
            const environment = new Environment({
                appDir: this.processRepresentation.appDir,
                appName: this.processRepresentation.appName,
                processName: this.processRepresentation.processName,
                pandoraLogsDir: LoggerBroker_1.getPandoraLogsDir()
            });
            pandora_env_1.EnvironmentUtil.getInstance().setCurrentEnvironment(environment);
        }
        // To start worker process monitoring
        MonitorManager_1.MonitorManager.injectProcessMonitor();
    }
    /**
     * A static method to handing the CLI
     */
    static cmd() {
        program
            .option('--params [params]')
            .parse(process.argv);
        let options;
        try {
            options = JSON.parse(program.params);
            assert(options.appName, 'The field appName required by ProcessBootstrap');
            assert(options.appDir, 'The field appDir required by ProcessBootstrap');
            assert(options.processName, 'The field processName required by ProcessBootstrap');
        }
        catch (err) {
            err.message = `Invalid options "${program.params}", ${err.message}`;
            LoggerBroker_1.consoleLogger.error(err);
            if (process.send) {
                process.send({ action: const_1.PROCESS_ERROR, error: err });
            }
            return;
        }
        // remove --params [params], some software dep on argv like node-red
        process.argv.pop();
        process.argv.pop();
        const processBootstrap = new ProcessBootstrap(options);
        process.on('message', (message) => {
            if (message.action === const_1.SHUTDOWN) {
                processBootstrap.stop().then(() => {
                    if (process.send) {
                        process.send({ action: const_1.FINISH_SHUTDOWN });
                    }
                }).catch(LoggerBroker_1.consoleLogger.error);
            }
        });
        const onProcessTerm = (sig) => {
            LoggerBroker_1.consoleLogger.info(`Process [name = ${options.processName}, pid = ${process.pid}] Receive a signal ${sig}, start to stopping, pid ${process.pid}`);
            processBootstrap.stop().then(() => {
                process.exit(0);
            }).catch((err) => {
                LoggerBroker_1.consoleLogger.error(err);
                process.exit(1);
            });
        };
        process.once('SIGQUIT', onProcessTerm.bind(null, 'SIGQUIT'));
        process.once('SIGTERM', onProcessTerm.bind(null, 'SIGTERM'));
        process.once('SIGINT', onProcessTerm.bind(null, 'SIGINT'));
        processBootstrap.start().then(() => {
            if (process.send) {
                process.send({ action: const_1.PROCESS_READY });
            }
        }).catch((err) => {
            LoggerBroker_1.consoleLogger.error(err);
            LoggerBroker_1.consoleLogger.error('An error occurred during the start of ProcessBootstrap.');
            if (process.send) {
                process.send({ action: const_1.PROCESS_ERROR, error: err });
            }
        });
    }
}
exports.ProcessBootstrap = ProcessBootstrap;
let cmdDid = false;
function cmd() {
    if (cmdDid) {
        return;
    }
    cmdDid = true;
    ProcessBootstrap.cmd();
}
exports.cmd = cmd;
// Handing CLI if this module be the main module
if (require.main === module || process.env.RUN_PROCESS_BOOTSTRAP_BY_FORCE) {
    delete process.env.RUN_PROCESS_BOOTSTRAP_BY_FORCE;
    cmd();
}
//# sourceMappingURL=ProcessBootstrap.js.map