'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const ApplicationHandler_1 = require("../application/ApplicationHandler");
const const_1 = require("../const");
const fs = require("fs");
const assert = require("assert");
const pandora_messenger_1 = require("pandora-messenger");
const LoggerBroker_1 = require("../universal/LoggerBroker");
const DaemonIntrospection_1 = require("./DaemonIntrospection");
const GlobalConfigProcessor_1 = require("../universal/GlobalConfigProcessor");
/**
 * Class Daemon
 */
class Daemon {
    constructor() {
        this.daemonLogger = LoggerBroker_1.getDaemonLogger();
        this.globalConfigProcesser = GlobalConfigProcessor_1.GlobalConfigProcessor.getInstance();
        this.globalConfig = this.globalConfigProcesser.getAllProperties();
        this.state = const_1.State.pending;
        this.apps = new Map();
    }
    /**
     * Start Daemon
     * @return {Promise<any>}
     */
    async start() {
        if (this.state === const_1.State.complete) {
            return;
        }
        return new Promise((resolve, reject) => {
            this.messengerServer = pandora_messenger_1.default.getServer({
                name: const_1.DAEMON_MESSENGER,
            });
            this.messengerServer.on(const_1.SEND_DAEMON_MESSAGE, (message, reply) => {
                this.handleCommand(message, reply);
            });
            this.messengerServer.ready(() => {
                return this.startMonitor().then(() => {
                    this.state = const_1.State.complete;
                    resolve();
                }).catch(err => {
                    this.daemonLogger.error(err);
                    process.exit(1);
                    reject(err);
                });
            });
            this.handleExit();
        });
    }
    /**
     * Handing all passive exit events, such as process signal, uncaughtException
     */
    handleExit() {
        process.on('uncaughtException', (err) => {
            this.daemonLogger.error(err);
            this.stop();
        });
        // SIGTERM AND SIGINT will trigger the exit event.
        ['SIGQUIT', 'SIGTERM', 'SIGINT'].forEach((sig) => {
            process.once(sig, () => {
                this.stop();
            });
        });
        process.once('exit', () => {
            this.stop();
        });
    }
    /**
     * Start an application
     *
     * @param {ApplicationRepresentation} applicationRepresentation
     * @returns {Promise<ApplicationHandler>}
     */
    async startApp(applicationRepresentation) {
        // require appName and appDir when app start
        const appName = applicationRepresentation.appName;
        const appDir = applicationRepresentation.appDir;
        assert(appName, `options.appName is required!`);
        assert(appDir, `options.appDir is required!`);
        assert(fs.existsSync(appDir), `${appDir} does not exists!`);
        assert(!this.apps.has(appName), `app[${appName}] has been initialized!`);
        const applicationHandler = new ApplicationHandler_1.ApplicationHandler(applicationRepresentation);
        await applicationHandler.start();
        this.apps.set(appName, applicationHandler);
        return applicationHandler;
    }
    /**
     * Reload an application
     * @param appName
     * @param processName
     * @return {Promise<ApplicationHandler>}
     */
    async reloadApp(appName, processName) {
        const handler = this.apps.get(appName);
        if (!handler) {
            throw new Error(`${appName} does not exists!`);
        }
        await handler.reload(processName);
        return handler;
    }
    /**
     * stop an application
     * @param appName
     * @return {Promise<ApplicationHandler>}
     */
    async stopApp(appName) {
        const handler = this.apps.get(appName);
        if (!handler) {
            throw new Error(`${appName} does not exists!`);
        }
        await handler.stop();
        this.apps.delete(appName);
        return handler;
    }
    /**
     * stop all the applications
     * @return {Promise<void>}
     */
    async stopAllApps() {
        for (const appName of this.apps.keys()) {
            await this.stopApp(appName);
        }
    }
    /**
     * stop an application
     * @return {Promise<void>}
     */
    async stop() {
        this.state = const_1.State.stopped;
        await this.stopAllApps();
        this.daemonLogger.info('daemon is going to stop');
        this.messengerServer.close();
        await this.stopMonitor();
        this.state = const_1.State.stopped;
    }
    /**
     * Start the monitor
     * @return {Promise<void>}
     */
    async startMonitor() {
        if (!this.monitor) {
            this.monitor = new this.globalConfig['monitor']();
        }
        return this.monitor.start();
    }
    /**
     * Stop the monitor
     * @return {Promise<void>}
     */
    async stopMonitor() {
        if (this.monitor) {
            return this.monitor.stop();
        }
    }
    /**
     * Handle daemon's command invocations
     * @param message
     * @param reply
     */
    handleCommand(message, reply) {
        const command = message.command;
        const args = message.args;
        switch (command) {
            case 'start':
                this.startApp(args).then(() => {
                    reply({ data: `${args.appName} started successfully! Run command [ pandora log ${args.appName} ] to get more information` });
                }).catch(err => {
                    reply({ error: `${args.appName} started failed, ${err && err.toString()}` });
                });
                break;
            case 'stopAll':
                this.stopAllApps().then(() => {
                    reply({ data: `all apps stopped successfully!` });
                }).catch(err => {
                    reply({ error: `all apps stopped failed, ${err && err.toString()}` });
                });
                break;
            case 'stopApp':
                this.stopApp(args.appName).then(() => {
                    reply({ data: `${args.appName} stopped successfully!` });
                }).catch(err => {
                    reply({ error: `${args.appName} stopped failed, ${err && err.toString()}` });
                });
                break;
            case 'restart':
                this.stopApp(args.appName).then(diedApp => {
                    return this.startApp(diedApp.appRepresentation);
                }).then(() => {
                    reply({ data: `${args.appName} restarted successfully!` });
                }).catch(err => {
                    reply({ error: `${args.appName} restarted failed, ${err && err.toString()}` });
                });
                break;
            case 'exit':
                this.stop().then(() => {
                    process.exit(0);
                });
                break;
            case 'list':
                const introspection = this.getIntrospection();
                return introspection.listApplication().then((data) => {
                    reply({ data });
                }).catch((error) => {
                    reply({ error });
                });
        }
    }
    getIntrospection() {
        if (!this.introspection) {
            this.introspection = new DaemonIntrospection_1.DaemonIntrospection(this);
        }
        return this.introspection;
    }
}
exports.Daemon = Daemon;
//# sourceMappingURL=Daemon.js.map